﻿{ The original file is: propvarutil.h, }
unit PropvarUtil;

interface

uses
  ActiveX, Windows, ShLwApi;

{$WARN SYMBOL_PLATFORM OFF}
function InitPropVariantFromBuffer(const pv: Pointer; cb: UINT;
  var ppropvar: TPropVariant): HRESULT; stdcall; external 'Propsys.dll' delayed;
function PropVariantClear(var pvar: TPropVariant): HRESULT; stdcall;
  external 'Ole32.dll' delayed;

function PropVariantToBooleanWithDefault(var propvarIn: TPropVariant;
  fDefault: Boolean): Boolean; stdcall; external 'Propsys.dll' delayed;
function PropVariantToInt16WithDefault(var propvarIn: TPropVariant;
  iDefault: Int16): Int16; stdcall; external 'Propsys.dll' delayed;
function PropVariantToUInt16WithDefault(var propvarIn: TPropVariant;
  uiDefault: UInt16): UInt16; stdcall; external 'Propsys.dll' delayed;
function PropVariantToInt32WithDefault(var propvarIn: TPropVariant;
  lDefault: Int32): Int32; stdcall; external 'Propsys.dll' delayed;
function PropVariantToUInt32WithDefault(var propvarIn: TPropVariant;
  ulDefault: UInt32): UInt32; stdcall; external 'Propsys.dll' delayed;
function PropVariantToInt64WithDefault(var propvarIn: TPropVariant;
  llDefault: Int64): Int64; stdcall; external 'Propsys.dll' delayed;
function PropVariantToUInt64WithDefault(var propvarIn: TPropVariant;
  ullDefault: UInt64): UInt64; stdcall; external 'Propsys.dll' delayed;
function PropVariantToDoubleWithDefault(var propvarIn: TPropVariant;
  dblDefault: Double): Double; stdcall; external 'Propsys.dll' delayed;
function PropVariantToStringWithDefault(var propvarIn: TPropVariant;
  pszDefault: LPCWSTR): PWideChar; stdcall; external 'Propsys.dll' delayed;
{$WARN SYMBOL_PLATFORM ON}
function InitPropVariantFromBoolean(fVal: Boolean;
  var ppropvar: TPropVariant): HRESULT;
function InitPropVariantFromInt16(nVal: Int16;
  var ppropvar: TPropVariant): HRESULT;
function InitPropVariantFromUInt16(uiVal: UInt16;
  var ppropvar: TPropVariant): HRESULT;
function InitPropVariantFromInt32(lVal: Int32;
  var ppropvar: TPropVariant): HRESULT;
function InitPropVariantFromUInt32(ulVal: UInt32;
  var ppropvar: TPropVariant): HRESULT;
function InitPropVariantFromInt64(llVal: Int64;
  var ppropvar: TPropVariant): HRESULT;
function InitPropVariantFromUInt64(ullVal: UInt64;
  var ppropvar: TPropVariant): HRESULT;
function InitPropVariantFromDouble(dblVal: Double;
  var ppropvar: TPropVariant): HRESULT;
function InitPropVariantFromString(sz: string;
  var ppropvar: TPropVariant): HRESULT;
function InitPropVariantFromGUIDAsBuffer(var guid: TGUID;
  var ppropvar: TPropVariant): HRESULT;

function IsPropVariantVector(var propvar: TPropVariant): Boolean;
function IsPropVariantString(var propvar: TPropVariant): Boolean;

type
  TPROPVAR_COMPARE_UNIT = (PVCU_DEFAULT = 0, PVCU_SECOND = 1, PVCU_MINUTE = 2,
    PVCU_HOUR = 3, PVCU_DAY = 4, PVCU_MONTH = 5, PVCU_YEAR = 6);
  TPROPVAR_COMPARE_FLAGS = (PVCF_DEFAULT = $00000000,
    // When comparing strings, use StrCmpLogical
    PVCF_TREATEMPTYASGREATERTHAN = $00000001,
    // Empty/null values are greater-than non-empty values
    PVCF_USESTRCMP = $00000002,  // When comparing strings, use StrCmp
    PVCF_USESTRCMPC = $00000004, // When comparing strings, use StrCmpC
    PVCF_USESTRCMPI = $00000008, // When comparing strings, use StrCmpI
    PVCF_USESTRCMPIC = $00000010 // When comparing strings, use StrCmpIC
    );

  {$WARN SYMBOL_PLATFORM OFF}
function PropVariantCompareEx(var propvar1, propvar2: TPropVariant;
  AUnit: TPROPVAR_COMPARE_UNIT; AFlags: TPROPVAR_COMPARE_FLAGS): Integer;
  stdcall; external 'Propsys.dll' delayed;
function PropVariantCompare(var propvar1, propvar2: TPropVariant): Integer;
{$WARN SYMBOL_PLATFORM ON}

implementation

procedure PropVariantInit(var pvar: TPropVariant);
begin
  ZeroMemory(@pvar, SizeOf(TPropVariant));
end;

function InitPropVariantFromBoolean(fVal: Boolean;
  var ppropvar: TPropVariant): HRESULT;
begin
  ppropvar.vt := VT_BOOL;
  if fVal then
    ppropvar.boolVal := True
  else
    ppropvar.boolVal := False;
  Result := S_OK;
end;

function InitPropVariantFromInt16(nVal: Int16;
  var ppropvar: TPropVariant): HRESULT;
begin
  ppropvar.vt := VT_I2;
  ppropvar.iVal := nVal;
  Result := S_OK;
end;

function InitPropVariantFromUInt16(uiVal: UInt16;
  var ppropvar: TPropVariant): HRESULT;
begin
  ppropvar.vt := VT_UI2;
  ppropvar.uiVal := uiVal;
  Result := S_OK;
end;

function InitPropVariantFromInt32(lVal: Int32;
  var ppropvar: TPropVariant): HRESULT;
begin
  ppropvar.vt := VT_I4;
  ppropvar.lVal := lVal;
  Result := S_OK;
end;

function InitPropVariantFromUInt32(ulVal: UInt32;
  var ppropvar: TPropVariant): HRESULT;
begin
  ppropvar.vt := VT_UI4;
  ppropvar.ulVal := ulVal;
  Result := S_OK;
end;

function InitPropVariantFromInt64(llVal: Int64;
  var ppropvar: TPropVariant): HRESULT;
begin
  ppropvar.vt := VT_I8;
  ppropvar.hVal.QuadPart := llVal;
  Result := S_OK;
end;

function InitPropVariantFromUInt64(ullVal: UInt64;
  var ppropvar: TPropVariant): HRESULT;
begin
  ppropvar.vt := VT_UI8;
  ppropvar.uhVal.QuadPart := ullVal;
  Result := S_OK;
end;

function InitPropVariantFromDouble(dblVal: Double;
  var ppropvar: TPropVariant): HRESULT;
begin
  ppropvar.vt := VT_R8;
  ppropvar.dblVal := dblVal;
  Result := S_OK;
end;

// Creates a VT_LPWSTR propvariant.
function InitPropVariantFromString(sz: string;
  var ppropvar: TPropVariant): HRESULT;
var
  hr: HRESULT;
begin
  ppropvar.vt := VT_LPWSTR;
  hr := SHStrDupW(PChar(sz), ppropvar.pwszVal);
  if FAILED(hr) then
  begin
    PropVariantInit(ppropvar);
  end;
  Result := hr;
end;

// Creates a VT_VECTOR | VT_UI1 propvariant.
function InitPropVariantFromGUIDAsBuffer(var guid: TGUID;
  var ppropvar: TPropVariant): HRESULT;
begin
  Result := InitPropVariantFromBuffer(@guid, SizeOf(TGUID), ppropvar);
end;

function IsPropVariantVector(var propvar: TPropVariant): Boolean;
begin
  Result := (propvar.vt and (VT_ARRAY or VT_VECTOR)) > 0;
end;

// If TRUE, propvar contains a unicode string.  Use PropVariantToStringWithDefault(propvar, NULL) to retrieve it.
function IsPropVariantString(var propvar: TPropVariant): Boolean;
begin
  Result := (PropVariantToStringWithDefault(propvar, nil) <> nil);
end;

function PropVariantCompare(var propvar1, propvar2: TPropVariant): Integer;
begin
  Result := PropVariantCompareEx(propvar1, propvar2, PVCU_DEFAULT,
    PVCF_DEFAULT);
end;

end.
