﻿// 2011/4/22 TJumplist.CreateJumpList 内で FAppIDが空の場合 SetAppID を実行しないように修正

unit uJumpList;

interface

uses
  Classes, SysUtils, Windows, ShlObj, Generics.Collections, ObjectArray,
  StrUtils;

type
  TTaskItem = class(TPersistent)
  private
    FArgument:     string;
    FDescription:  string;
    FIconIndex:    Integer;
    FIconLocation: string;
    FPath:         string;
    FTitle:        string;
    procedure SetArgument(const Value: string);
    procedure SetDescription(const Value: string);
    procedure SetIconIndex(const Value: Integer);
    procedure SetIconLocation(const Value: string);
    procedure SetPath(const Value: string);
    procedure SetTitle(const Value: string);
  protected
    function CreateSeparatorLink(out AShellLink: IShellLink): Boolean;
    function CreateShellItem(out AShellItem: IShellItem): Boolean;
    function CreateShellLink(out AShellLink: IShellLink): Boolean;
  public
    constructor Create; overload;
    constructor Create(APath, ATitle, AArgument: string); overload;

    procedure Assign(Source: TPersistent); override;

    function CreateObject(out AShellObject: IUnknown): Boolean;
    property Argument: string read FArgument write SetArgument;
    property Description: string read FDescription write SetDescription;
    property IconIndex: Integer read FIconIndex write SetIconIndex;
    property IconLocation: string read FIconLocation write SetIconLocation;
    property Path: string read FPath write SetPath;
    property Title: string read FTitle write SetTitle;
  end;

  TTaskList = TObjectList<TTaskItem>;

  TJumplist = class
  private
    FAppID:      string;
    FCategories: TObjectDictionary<string, TTaskList>;
    FTasks:      TTaskList;
    function AddCategoryToList(ACategoryName: string; var oaRemoved: IObjectArray;
      List: TTaskList; var CDL: ICustomDestinationList): Boolean;
    function AddTasksToList(List: TTaskList;
      var CDL: ICustomDestinationList): Boolean;
    function GetCategories(ACategory: string): TTaskList;
    function IsItemInArray(ASI: IShellItem; var oaRemoved: IObjectArray)
      : Boolean; overload;
    function IsItemInArray(Item: TTaskItem; var oaRemoved: IObjectArray)
      : Boolean; overload;
  public
    constructor Create(const AAppID: string);
    destructor Destroy; override;
    function CreateJumpList: Boolean;
    procedure DeleteJumpList;
    property AppID: string read FAppID;
    property Categories[ACategory: string]: TTaskList read GetCategories;
    property Tasks: TTaskList read FTasks;
  end;

implementation

uses
  ActiveX, PropSys, PropKey, PropVarUtil;

function PathFromIDList(pidl: PItemIDList): string;
var
  Buff: PWideChar;
begin
  Buff := AllocMem(MAX_PATH + 1);
  SHGetPathFromIDList(pidl, Buff);
  Result := Buff;
  FreeMem(Buff);
end;

{ TTaskItem }

constructor TTaskItem.Create;
begin

end;

procedure TTaskItem.Assign(Source: TPersistent);
var
  Src: TTaskItem;
begin
  if Source is TTaskItem then
  begin
    Src := Source as TTaskItem;
    FArgument := Src.FArgument;
    FDescription := Src.FDescription;
    FIconIndex := Src.FIconIndex;
    FIconLocation := Src.FIconLocation;
    FPath := Src.FPath;
    FTitle := Src.FTitle;
  end else
    inherited;
end;

constructor TTaskItem.Create(APath, ATitle, AArgument: string);
begin
  FPath := APath;
  FTitle := ATitle;
  FArgument := AArgument;
end;

function TTaskItem.CreateObject(out AShellObject: IUnknown): Boolean;
var
  SL: IShellLink;
  SI: IShellItem;
begin
  AShellObject := nil;

  if FArgument <> '' then
  begin
    // 引数が空で無ければ IShellLink を作成
    if CreateShellLink(SL) then
    begin
      AShellObject := SL as IUnknown;
    end;
  end
  else if FPath <> '' then
  begin
    // ファイルパスが設定されてれば IShellItem を作成
    if CreateShellItem(SI) then
    begin
      AShellObject := SI as IUnknown;
    end;
  end else begin
    // どちらも空の場合はセパレータを作成
    if CreateSeparatorLink(SL) then
    begin
      AShellObject := SL as IUnknown;
    end;
  end;
  Result := Assigned(AShellObject);
end;

function TTaskItem.CreateSeparatorLink(out AShellLink: IShellLink): Boolean;
var
  PS:      IPropertyStore;
  Propvar: TPropVariant;
begin
  Result := False;
  if Succeeded(CoCreateInstance(CLSID_ShellLink, nil, CLSCTX_INPROC_SERVER,
      IID_IPropertyStore, PS)) then
  begin
    if Succeeded(InitPropVariantFromBoolean(True, Propvar)) then
    begin
      if Succeeded(PS.SetValue(PKEY_AppUserModel_IsDestListSeparator,
          Propvar)) then
      begin
        if Succeeded(PS.Commit) then
        begin
          Result := Succeeded(PS.QueryInterface(IID_IShellLink, AShellLink));
        end;
      end;
      PropVariantClear(Propvar);
    end;
  end;
end;

function TTaskItem.CreateShellItem(out AShellItem: IShellItem): Boolean;
var
  Desktop:     IShellFolder;
  pidl:        PItemIDList;
  chEaten:     ULONG;
  dwAttribute: ULONG;
begin
  Result := False;
  if Succeeded(SHGetDesktopFolder(Desktop)) then
  begin
    if Succeeded(Desktop.ParseDisplayName(0, nil, PChar(FPath), chEaten, pidl,
        dwAttribute)) then
    begin
      Result := Succeeded(SHCreateItemFromIDList(pidl, IID_IShellItem,
          AShellItem));
      CoTaskMemFree(pidl);
    end;
  end;
end;

function TTaskItem.CreateShellLink(out AShellLink: IShellLink): Boolean;
var
  HR:      HRESULT;
  SL:      IShellLink;
  PS:      IPropertyStore;
  Propvar: TPropVariant;
begin
  Result := False;

  HR := E_FAIL;
  if Succeeded(CoCreateInstance(CLSID_ShellLink, nil, CLSCTX_INPROC_SERVER,
      IID_IShellLink, SL)) then
  begin
    if Failed(SL.SetPath(PChar(FPath))) then
      Exit;

    if Failed(SL.SetArguments(PChar(FArgument))) then
      Exit;

    if (FDescription <> '') and
      Failed(SL.SetDescription(PChar(FDescription))) then
      Exit;

    if (FIconLocation <> '') and Failed(SL.SetIconLocation(PChar(FIconLocation),
        FIconIndex)) then
      Exit;

    if Succeeded(SL.QueryInterface(IID_IPropertyStore, PS)) then
    begin
      if Succeeded(InitPropVariantFromString(FTitle, Propvar)) then
      begin
        if Succeeded(PS.SetValue(PKEY_Title, Propvar)) then
        begin
          HR := PS.Commit;
        end;
        PropVariantClear(Propvar);
      end;
    end;
    if Succeeded(HR) then
    begin
      Result := Succeeded(SL.QueryInterface(IID_IShellLink, AShellLink));
    end;
  end;
end;

procedure TTaskItem.SetArgument(const Value: string);
begin
  FArgument := Value;
end;

procedure TTaskItem.SetDescription(const Value: string);
begin
  FDescription := Value;
end;

procedure TTaskItem.SetIconIndex(const Value: Integer);
begin
  FIconIndex := Value;
end;

procedure TTaskItem.SetIconLocation(const Value: string);
begin
  FIconLocation := Value;
end;

procedure TTaskItem.SetPath(const Value: string);
begin
  FPath := Value;
end;

procedure TTaskItem.SetTitle(const Value: string);
begin
  FTitle := Value;
end;

constructor TJumplist.Create(const AAppID: string);
begin
  FCategories := TObjectDictionary<string, TTaskList>.Create([doOwnsValues]);
  FTasks := TTaskList.Create(True);
  FAppID := AAppID;
end;

destructor TJumplist.Destroy;
begin
  FreeAndNil(FTasks);
  FreeAndNil(FCategories);
  inherited;
end;

{ TJumplist }

function TJumplist.AddCategoryToList(ACategoryName: string;
  var oaRemoved: IObjectArray; List: TTaskList;
  var CDL: ICustomDestinationList): Boolean;
var
  OC:   IObjectCollection;
  Item: TTaskItem;

  Idx: Integer;

  OA: IObjectArray;
  UK: IUnknown;
begin
  Result := False;
  if Succeeded(CoCreateInstance(CLSID_EnumerableObjectCollection, nil,
      CLSCTX_INPROC, IID_IObjectCollection, OC)) then
  begin
//    for Item in List do
    for Idx := 0 to List.Count - 1 do
    begin
      Item := List.Items[Idx];
      if not IsItemInArray(Item, oaRemoved) then
      begin
        if Item.CreateObject(UK) then
          OC.AddObject(UK);
      end;
    end;

    if Succeeded(OC.QueryInterface(IID_IObjectArray, OA)) then
    begin
      Result := Succeeded(CDL.AppendCategory(PChar(ACategoryName), OA));
    end;
  end;
end;

function TJumplist.AddTasksToList(List: TTaskList;
  var CDL: ICustomDestinationList): Boolean;
var
  OC:   IObjectCollection;
  Item: TTaskItem;

  OA: IObjectArray;
  UK: IUnknown;
begin
  Result := False;
  if Succeeded(CoCreateInstance(CLSID_EnumerableObjectCollection, nil,
      CLSCTX_INPROC, IID_IObjectCollection, OC)) then
  begin
    for Item in List do
    begin
      if Item.CreateObject(UK) then
        OC.AddObject(UK);
    end;

    if Succeeded(OC.QueryInterface(IID_IObjectArray, OA)) then
    begin
      Result := Succeeded(CDL.AddUserTasks(OA));
    end;
  end;
end;

function TJumplist.CreateJumpList: Boolean;
var
  CDL:       ICustomDestinationList;
  MinSlots:  UINT;
  oaRemoved: IObjectArray;

  Item: TPair<string, TTaskList>;
  Done: Boolean;
begin
  Result := False;

  if Failed(CoCreateInstance(CLSID_DestinationList, nil, CLSCTX_INPROC_SERVER,
      IID_ICustomDestinationList, CDL)) then
    Exit;

  if (FAppID <> '') And Failed(CDL.SetAppID(PChar(FAppID))) then
    Exit;

  // ジャンプリストの編集開始
  if Failed(CDL.BeginList(MinSlots, IID_IObjectArray, oaRemoved)) then
    Exit;

  Done := True;
  // カスタムカテゴリの追加
  for Item in FCategories do
  begin
    if not AddCategoryToList(Item.Key, oaRemoved, Item.Value, CDL) then
      Done := False;
  end;

  // タスクの追加
  if Done then
  begin
    Done := AddTasksToList(Tasks, CDL);
  end;

  // 変更の適用又は破棄
  if Done then
    Result := Succeeded(CDL.CommitList)
  else
    Result := Succeeded(CDL.AbortList);
end;

procedure TJumplist.DeleteJumpList;
var
  CDL: ICustomDestinationList;
begin
  if Succeeded(CoCreateInstance(CLSID_DestinationList, nil,
      CLSCTX_INPROC_SERVER, IID_ICustomDestinationList, CDL)) then
  begin
    CDL.DeleteList(nil);
  end;
end;

function TJumplist.GetCategories(ACategory: string): TTaskList;
begin
  if not FCategories.ContainsKey(ACategory) then
  begin
    FCategories.Add(ACategory, TTaskList.Create(True));
  end;
  Result := FCategories.Items[ACategory];
end;

function TJumplist.IsItemInArray(ASI: IShellItem;
  var oaRemoved: IObjectArray): Boolean;
var
  Idx:       Integer;
  Count:     UINT;
  siCompare: IShellItem;
  Order:     Integer;
begin
  Result := False;
  if Succeeded(oaRemoved.GetCount(Count)) then
  begin
    for Idx := 0 to Count - 1 do
    begin
      if Succeeded(oaRemoved.GetAt(Idx, IID_IShellItem, siCompare)) then
      begin
        Result := Succeeded(siCompare.Compare(ASI, SICHINT_CANONICAL, Order))
          and (0 = Order);
        siCompare := nil;
      end;
    end;
  end;
end;

function TJumplist.IsItemInArray(Item: TTaskItem;
  var oaRemoved: IObjectArray): Boolean;
var
  Idx:   Integer;
  Count: UINT;
  SI:    IShellItem;
  SL:    IShellLink;
  PS:    IPropertyStore;

  pidl:     PItemIDList;
  Propvar:  TPropVariant;
  Path:     string;
  Argument: string;
begin
  Result := False;
  if Succeeded(oaRemoved.GetCount(Count)) then
  begin
    for Idx := 0 to Count - 1 do
    begin
      Argument := '';
      Path := '';
      if Succeeded(oaRemoved.GetAt(Idx, IID_IShellItem, SI)) then
      begin
        if Succeeded(SHGetIDListFromObject(SI, pidl)) then
        begin
          Path := PathFromIDList(pidl);
          CoTaskMemFree(pidl);
        end;
      end
      else if Succeeded(oaRemoved.GetAt(Idx, IID_IShellLink, SL)) then
      begin
        if Succeeded(SL.QueryInterface(IID_IPropertyStore, PS)) then
        begin
          if Succeeded(PS.GetValue(PKEY_Link_Arguments, Propvar)) then
          begin
            Path := PropVariantToStringWithDefault(Propvar, nil);
            PropVariantClear(Propvar);
          end;
          if Succeeded(SL.GetIDList(pidl)) then
          begin
            if Assigned(pidl) then
            begin
              Path := PathFromIDList(pidl);
              CoTaskMemFree(pidl);
            end else
              Path := '';
          end;
        end;
      end;
      if (Argument <> '') and (Path <> '') then
        Result := AnsiSameText(Item.Path, Path) and
          AnsiSameText(Item.Argument, Argument)
      else if Path <> '' then
        Result := AnsiSameText(Item.Path, Path);
      if Result then
        break;
    end;
  end;
end;

end.
